from ..DynamicDiffusionModel import DynamicDiffusionModel
import numpy as np
import networkx as nx
import future.utils
from scipy import stats


__author__ = "Letizia Milli"
__license__ = "BSD-2-Clause"
__email__ = "letizia.milli@isti.cnr.it"


class DynKerteszThresholdModel(DynamicDiffusionModel):
    """
             Node Parameters to be specified via ModelConfig

            :param profile: The node profile. As default a value of 0.1 is assumed for all nodes.
         """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
            "Blocked": -1
        }

        self.parameters = {
            "model": {
                "adopter_rate": {
                    "descr": "Exogenous adoption rate",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "percentage_blocked": {
                    "descr": "Percentage of blocked nodes",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                }
            },
            "nodes": {
                "threshold": {
                    "descr": "Node threshold",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                }
            },
            "edges": {},
        }

        self.name = "Kertesz Threhold"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        # streaming
        if self.stream_execution:
            raise ValueError("Streaming network not allowed.")
        # snapshot
        else:
            if self.actual_iteration == 0:

                if min(actual_status.values()) == 0:
                    number_node_blocked = int(float(self.graph.number_of_nodes()) *
                                              float(self.params['model']['percentage_blocked']))

                    i = 0
                    while i < number_node_blocked:
                        # select a random node
                        node = list(self.graph.nodes())[np.random.randint(0, self.graph.number_of_nodes())]

                        # node not infected
                        if actual_status[node] == 0:
                            # node blocked
                            actual_status[node] = -1
                            self.status[node] = -1
                            i += 1

                self.actual_iteration += 1
                delta, node_count, status_delta = self.status_delta(actual_status)
                if node_status:
                    return {"iteration": 0, "status": actual_status.copy(),
                            "node_count": node_count.copy(), "status_delta": status_delta.copy()}
                else:
                    return {"iteration": 0, "status": {},
                            "node_count": node_count.copy(), "status_delta": status_delta.copy()}

            for node in self.graph.nodes():
                if self.status[node] == 0:
                    if self.params['model']['adopter_rate'] > 0:
                        xk = (0, 1)
                        pk = (1 - self.params['model']['adopter_rate'], self.params['model']['adopter_rate'])
                        probability = stats.rv_discrete(name='probability', values=(xk, pk))
                        number_probability = probability.rvs()

                        if number_probability == 1:
                            actual_status[node] = 1
                            continue

                    neighbors = list(self.graph.neighbors(node))
                    if len(neighbors) == 0:
                        continue

                    if isinstance(self.graph, nx.DiGraph):
                        neighbors = self.graph.predecessors(node)

                    infected = 0
                    for v in neighbors:
                        if self.status[v] != -1:
                            infected += self.status[v]

                    infected_ratio = float(infected) / len(neighbors)
                    if infected_ratio >= self.params['nodes']['threshold'][node]:
                        actual_status[node] = 1

            delta, node_count, status_delta = self.status_delta(actual_status)
            self.status = actual_status
            self.actual_iteration += 1

            if node_status:
                return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": self.actual_iteration - 1, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}


from ..DiffusionModel import DiffusionModel
import numpy as np

__author__ = "Alina Sirbu"
__email__ = "alina.sirbu@unipi.it"


class MajorityRuleModel(DiffusionModel):
    """

    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
        }

        self.parameters = {"model": {
            "q": {
                "descr": "Number of randomly chosen voters",
                "range": [0, len(self.graph.nodes)],
                "optional": False
            }
        },
            "nodes": {},
            "edges": {}
        }

        self.name = "Majority Rule"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """

        # One iteration changes the opinion of at most q voters using the following procedure:
        # - select randomly q voters
        # - compute majority opinion
        # - if tie all agents take opinion +1
        # - if not tie, all agents take majority opinion

        self.clean_initial_status(self.available_statuses.values())

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(self.status)
            if node_status:
                return {"iteration": 0, "status": self.status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        # select q random nodes
        discussion_group = [list(self.graph.nodes)[i]
                            for i in np.random.randint(0, self.graph.number_of_nodes(), self.params['model']['q'])]

        # compute majority
        majority_vote = 1
        vote_sum = sum([self.status[node] for node in discussion_group])
        if vote_sum < (self.params["model"]["q"] / 2.0):
            majority_vote = 0  # in case of tie, majority_vote remains 1

        # update status of nodes in discussion group
        delta = {}
        status_delta = {st: 0 for st in self.available_statuses.values()}

        for listener in discussion_group:
            if majority_vote != self.status[listener]:
                delta[listener] = majority_vote

                status_delta[self.status[listener]] += 1
                for x in self.available_statuses.values():
                    if x != self.status[listener]:
                        status_delta[x] -= 1

            self.status[listener] = majority_vote

        # fix
        node_count = {st: len([n for n in self.status if self.status[n] == st])
                      for st in self.available_statuses.values()}

        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}





from ndlib.models.compartments.Compartment import Compartiment

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class CountDown(Compartiment):

    def __init__(self, name, iterations, **kwargs):
        super(self.__class__, self).__init__(kwargs)
        self.iterations = iterations
        self.name = name

    def execute(self, node, graph, status, status_map, *args, **kwargs):
        if self.name in graph.nodes[node]:
            graph.nodes[node][self.name] -= 1
        else:
            graph.nodes[node][self.name] = self.iterations

        test = graph.nodes[node][self.name] == 0
        if test:
            return self.compose(node, graph, status, status_map, kwargs)

        return False



from ..DiffusionModel import DiffusionModel
import numpy as np

__author__ = "Alina Sirbu"
__email__ = "alina.sirbu@unipi.it"


class SznajdModel(DiffusionModel):
    """

    """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
         }

        self.name = "Sznajd"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        # One iteration changes the opinion of several voters using the following procedure:
        # - select randomly one voter (speaker 1)
        # - select randomly one of its neighbours (speaker 2)
        # - if the two voters agree, their neighbours take their opinion

        self.clean_initial_status(self.available_statuses.values())

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(self.status)
            if node_status:
                return {"iteration": 0, "status": self.status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        delta = {}
        status_delta = {st: 0 for st in self.available_statuses.values()}

        # select a random node
        speaker1 = list(self.graph.nodes)[np.random.randint(0, self.graph.number_of_nodes())]

        # select a random neighbour
        neighbours = list(self.graph.neighbors(speaker1))
        if self.graph.directed:
            # add also the predecessors
            neighbours += list(self.graph.predecessors(speaker1))

        speaker2 = neighbours[np.random.randint(0, len(neighbours))]

        if self.status[speaker1] == self.status[speaker2]:
            # select listeners (all neighbours of two speakers)
            neighbours = list(self.graph.neighbors(speaker1)) + list(self.graph.neighbors(speaker2))

            if self.graph.directed:
                # assumed if a->b then b can be influenced by a
                # but not the other way around - the link between the speakers doesn't matter
                neighbours = list(self.graph.successors(speaker1)) + list(self.graph.successors(speaker2))

            # update status of listeners
            for listener in neighbours:
                if self.status[speaker1] != self.status[listener]:
                    delta[listener] = self.status[speaker1]
                    status_delta[self.status[listener]] += 1
                    for x in self.available_statuses.values():
                        if x != self.status[listener]:
                            status_delta[x] -= 1

                self.status[listener] = self.status[speaker1]

        node_count = {st: len([n for n in self.status if self.status[n] == st])
                      for st in self.available_statuses.values()}

        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}





from ..DiffusionModel import DiffusionModel
import numpy as np
from scipy import stats
import future.utils

__author__ = "Letizia Milli"
__license__ = "BSD-2-Clause"
__email__ = "letizia.milli@di.unipi.it"


class KerteszThresholdModel(DiffusionModel):
    """
         Node/Model Parameters to be specified via ModelConfig

        :param threshold: The node threshold. As default a value of 0.1 is assumed for all nodes.
        :param adopter_rate: The probability of spontaneous adoptions. Defaults value 0.
        :param fraction_infected: The percentage of blocked nodes. Default value 0.1.
     """

    def __init__(self, graph, seed=None):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph, seed)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
            "Blocked": -1
        }

        self.parameters = {
            "model": {
                "adopter_rate": {
                    "descr": "Exogenous adoption rate",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0
                },
                "percentage_blocked": {
                    "descr": "Percentage of blocked nodes",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                }
            },
            "nodes": {
                "threshold": {
                    "descr": "Node threshold",
                    "range": [0, 1],
                    "optional": True,
                    "default": 0.1
                }
            },
            "edges": {},
        }

        self.name = "Kertesz Threhold"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:

            if min(actual_status.values()) == 0:
                number_node_blocked = int(float(self.graph.number_of_nodes()) *
                                          float(self.params['model']['percentage_blocked']))

                i = 0
                while i < number_node_blocked:
                    # select a random node
                    #print("aa", len(self.graph.nodes()))#, self.graph.number_of_nodes(), np.random.randint(0, self.graph.number_of_nodes()))
                    try:
                        node = list(self.graph.nodes())[np.random.randint(0, self.graph.number_of_nodes())]

                        # node not infected
                        if actual_status[node] == 0:

                            # node blocked
                            actual_status[node] = -1
                            self.status[node] = -1
                            i += 1
                    except:
                        i += 1

            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        for node in self.graph.nodes:
            if self.status[node] == 0:
                if self.params['model']['adopter_rate'] > 0:
                    xk = (0, 1)
                    pk = (1-self.params['model']['adopter_rate'], self.params['model']['adopter_rate'])
                    probability = stats.rv_discrete(name='probability', values=(xk, pk))
                    number_probability = probability.rvs()

                    if number_probability == 1:
                        actual_status[node] = 1
                        continue

                neighbors = list(self.graph.neighbors(node))
                if len(neighbors) == 0:
                    continue

                if self.graph.directed:
                    neighbors = self.graph.predecessors(node)

                infected = 0
                for v in neighbors:
                    if self.status[v] != -1:
                        infected += self.status[v]

                infected_ratio = float(infected)/len(neighbors)
                if infected_ratio >= self.params['nodes']['threshold'][node]:
                    actual_status[node] = 1

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}



from ndlib.models.DiffusionModel import DiffusionModel
import numpy as np
import future.utils
import networkx as nx

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class ICEModel(DiffusionModel):
    """
        Parameter free model: probability of diffusion tied to community embeddedness of individual nodes
    """

    def __init__(self, graph):
        """
             Model Constructor

             :param graph: A networkx graph object
         """
        super(self.__class__, self).__init__(graph)
        self.available_statuses = {
            "Susceptible": 0,
            "Infected": 1,
            "Removed": 2
        }

        self.parameters = {
            "model": {},
            "nodes": {},
            "edges": {}
        }

        self.name = "Community Embeddedness"

    def iteration(self, node_status=True):
        """
        Execute a single model iteration

        :return: Iteration_id, Incremental node status (dictionary node->status)
        """
        self.clean_initial_status(self.available_statuses.values())
        actual_status = {node: nstatus for node, nstatus in future.utils.iteritems(self.status)}

        if self.actual_iteration == 0:
            self.actual_iteration += 1
            delta, node_count, status_delta = self.status_delta(actual_status)
            if node_status:
                return {"iteration": 0, "status": actual_status.copy(),
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}
            else:
                return {"iteration": 0, "status": {},
                        "node_count": node_count.copy(), "status_delta": status_delta.copy()}

        edge_embeddedness = {}
        for u in self.graph.nodes:
            if self.status[u] != 1:
                continue

            edge_embeddedness[u] = {}
            com_embeddedness = {}

            neighbors = list(self.graph.neighbors(u))  # neighbors and successors (in DiGraph) produce the same result
            embeddedness = 1
            for v in neighbors:
                same_community_neighbors = 0
                neighbors_v = list(self.graph.neighbors(v))
                for neighbor in neighbors_v:
                    if neighbor in neighbors and self.params['nodes']['com'][neighbor] == self.params['nodes']['com'][u]:
                        same_community_neighbors += 1
                tmp_embeddedness = float(same_community_neighbors) / float(
                    len(neighbors) + len(neighbors_v) - same_community_neighbors)
                edge_embeddedness[u][v] = tmp_embeddedness
                if tmp_embeddedness < embeddedness:
                    embeddedness = tmp_embeddedness
            com_embeddedness[self.params['nodes']['com'][u]] = embeddedness


           # neighbors = list(self.graph.neighbors(u))  # neighbors and successors (in DiGraph) produce the same result
            #same_community_neighbors = [n for n in neighbors if self.params['nodes']['com'][u] == self.params['nodes']['com'][n]]

            # Standard threshold
            if len(neighbors) > 0:

                for v in neighbors:
                    if actual_status[v] == 0:

                        if self.params['nodes']['com'][u] == self.params['nodes']['com'][v]:
                            threshold = edge_embeddedness[u][v]

                        else: # across communities
                            threshold = com_embeddedness[self.params['nodes']['com'][u]]

                        flip = np.random.random_sample()
                        if flip <= threshold:
                            actual_status[v] = 1

            actual_status[u] = 2

        delta, node_count, status_delta = self.status_delta(actual_status)
        self.status = actual_status
        self.actual_iteration += 1

        if node_status:
            return {"iteration": self.actual_iteration - 1, "status": delta.copy(),
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}
        else:
            return {"iteration": self.actual_iteration - 1, "status": {},
                    "node_count": node_count.copy(), "status_delta": status_delta.copy()}



from ndlib.models.compartments.Compartment import Compartiment
import networkx as nx
import numpy as np
import operator

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class NodeNumericalAttribute(Compartiment):

    def __init__(self, attribute, value=None, op=None, probability=1, **kwargs):
        super(self.__class__, self).__init__(kwargs)
        self.__available_operators = {"==": operator.__eq__, "<": operator.__lt__,
                                      ">": operator.__gt__, "<=": operator.__le__,
                                      ">=": operator.__ge__, "!=": operator.__ne__,
                                      "IN": (operator.__ge__, operator.__le__)}

        self.attribute = attribute
        self.attribute_range = value
        self.probability = probability
        self.operator = op

        if self.attribute_range is None:
            raise ValueError("A valid attribute value must be provided")

        if self.operator is not None and self.operator in self.__available_operators:
            if self.operator == "IN":
                if not isinstance(self.attribute_range, list) or self.attribute_range[1] < self.attribute_range[0]:
                    raise ValueError("A range list is required to test IN condition")
            else:
                if not isinstance(self.attribute_range, int):
                    if not isinstance(self.attribute_range, float):
                        raise ValueError("A numeric value is required to test the selected condition")
        else:
            raise ValueError("The operator provided '%s' is not valid" % operator)

    def execute(self, node, graph, status, status_map, *args, **kwargs):

        val = nx.get_node_attributes(graph, self.attribute)[node]
        p = np.random.random_sample()

        if self.operator == "IN":
            condition = self.__available_operators[self.operator][0](val, self.attribute_range[0]) and \
                        self.__available_operators[self.operator][1](val, self.attribute_range[1])
        else:
            condition = self.__available_operators[self.operator](val, self.attribute_range)

        test = condition and p <= self.probability

        if test:
            return self.compose(node, graph, status, status_map, kwargs)

        return False



import os
import matplotlib as mpl
if os.environ.get('DISPLAY', '') == '':
    print('no display found. Using non-interactive Agg backend')
    mpl.use('Agg')
import matplotlib.pyplot as plt
import future.utils


__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class OpinionEvolution(object):

    def __init__(self, model, trends):
        """
        :param model: The model object
        :param trends: The computed simulation trends
        """
        self.model = model
        self.srev = trends
        self.ylabel = "Opinion"

    def plot(self, filename=None):
        """
        Generates the plot

        :param filename: Output filename
        :param percentile: The percentile for the trend variance area
        """

        descr = ""
        infos = self.model.get_info()

        for k, v in future.utils.iteritems(infos):
            descr += "%s: %s, " % (k, v)
        descr = descr[:-2].replace("_", " ")

        nodes2opinions = {}
        node2col = {}

        last_it = self.srev[-1]['iteration'] + 1
        last_seen = {}

        for it in self.srev:
            sts = it['status']
            its = it['iteration']
            for n, v in sts.items():
                if n in nodes2opinions:
                    last_id = last_seen[n]
                    last_value = nodes2opinions[n][last_id]

                    for i in range(last_id, its):
                        nodes2opinions[n][i] = last_value

                    nodes2opinions[n][its] = v
                    last_seen[n] = its
                else:
                    nodes2opinions[n] = [0]*last_it
                    nodes2opinions[n][its] = v
                    last_seen[n] = 0
                    if v < 0.33:
                        node2col[n] = '#ff0000'
                    elif 0.33 <= v <= 0.66:
                        node2col[n] = '#00ff00'
                    else:
                        node2col[n] = '#0000ff'

        mx = 0
        for k, l in future.utils.iteritems(nodes2opinions):
            if mx < last_seen[k]:
                mx = last_seen[k]
            x = list(range(0, last_seen[k]))
            y = l[0:last_seen[k]]
            plt.plot(x, y, lw=1, alpha=0.5, color=node2col[k])

        plt.title(descr)
        plt.xlabel("Iterations", fontsize=24)
        plt.ylabel(self.ylabel, fontsize=24)
        plt.legend(loc="best", fontsize=18)

        plt.tight_layout()
        if filename is not None:
            plt.savefig(filename)
            plt.clf()
        else:
            plt.show()


try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import sys
import json
import os.path
import re
import networkx as nx
from ndlib.models.CompositeModel import CompositeModel

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


class ExperimentParser(object):

    def __init__(self):
        self.imports = 'import networkx as nx\n' \
                       'import numpy as np\n' \
                       'import json\n' \
                       'from ndlib.models.ModelConfig import Configuration\n' \
                       'from ndlib.models.CompositeModel import CompositeModel\n' \
                       'from ndlib.models.compartments.NodeStochastic import NodeStochastic\n' \
                       'from ndlib.models.compartments.NodeThreshold import NodeThreshold\n' \
                       'from ndlib.models.compartments.NodeCategoricalAttribute import NodeCategoricalAttribute\n' \
                       'from ndlib.models.compartments.NodeNumericalAttribute import NodeNumericalAttribute\n' \
                       'from ndlib.models.compartments.EdgeStochastic import EdgeStochastic\n' \
                       'from ndlib.models.compartments.EdgeCategoricalAttribute import EdgeCategoricalAttribute\n' \
                       'from ndlib.models.compartments.EdgeNumericalAttribute import EdgeNumericalAttribute\n' \
                       'from ndlib.models.compartments.ConditionalComposition import ConditionalComposition\n' \
                       'from ndlib.models.compartments.CountDown import CountDown\n'

        self.script = ""
        self.starting = ('STATUS', 'MODEL', 'COMPARTMENT', 'RULE', 'IF', 'INITIALIZE',
                         'CREATE_NETWORK', 'LOAD_NETWORK', 'EXECUTE')
        self.__model_name = None
        self.__net_name = None
        self.query = None
        self.__statuses = {}
        self.__compartments = {}
        self.model = CompositeModel(nx.Graph())

    def read_query_file(self, filename):
        with open(filename) as f:
            self.query = f.read()

    def set_query(self, query):
        self.query = query

    def parse(self):

        # Tokenizing directives
        identified_directives = {}
        cmd = []
        bucket = []
        lines = self.query.split("\n")
        for line in lines:
            # Comment handling
            if len(line) == 0 or line[0] == "#":
                continue

            line = self.__sanitize_string(line)
            kwd = line.rstrip().split(" ")[0]

            if kwd in self.starting:
                identified_directives[kwd] = None
                if len(bucket) > 0:
                    cmd.append(bucket)
                bucket = [line.rstrip()]
            else:
                line = line.rstrip()
                if len(line) > 0:
                    bucket.append(line.rstrip())
        cmd.append(bucket)

        # Query consistency checks
        self.__check_components(identified_directives)

        # Lookup directives: script generation and execution
        for statement in cmd:
            key = statement[0].rstrip().split(" ")[0]

            if key == "MODEL":
                code = self.__model_creation(statement)

            elif key == "STATUS":
                code = self.__status_definition(statement)

            elif key == "COMPARTMENT":
                code = self.__compartment_definition(statement)

            elif key == "RULE":
                code = self.__rule_definition(statement)

            elif key == "IF":
                code = self.__conditional_compartment_composition(statement)

            elif key == "INITIALIZE":
                code = self.__model_configuration(statement)

            elif key == "CREATE_NETWORK":
                code = self.__network_generation(statement)

            elif key == "LOAD_NETWORK":
                code = self.__network_loading(statement)

            elif key == "EXECUTE":
                code = self.__execution_statement(statement)

            else:
                raise ValueError("The keyword '%s' is not defined: check your syntax" % key)
            self.script = "%s%s" % (self.script, code)

        self.__clean_imports()
        self.script = "%s\n%s" % (self.imports, self.script)

    def execute_query(self):
        # Query execution
        old_stdout = sys.stdout
        redirected_output = sys.stdout = StringIO()

        try:
            exec(self.script, locals(), globals())
        except SyntaxError:
            raise ValueError("Experiment description malformed (Incorrect statement ordering): check your syntax")

        sys.stdout = old_stdout
        result = json.loads(redirected_output.getvalue())
        trends = self.model.build_trends(result)
        trends[0]['Statuses'] = {str(v): k for k, v in self.model.available_statuses.items()}
        return trends

    def __clean_imports(self):

        libs = self.imports.split("\n")
        new_imports = "\n".join(libs[:5])
        compartments = set(self.__compartments.values())
        rs = ["\\b%s\\b " % x for x in compartments]
        cps = r"|".join(rs)

        for lib in libs[5:]:
            r = re.compile(cps, flags=re.I | re.X)
            match = r.findall(lib)
            if match:
                new_imports += "\n%s\n" % lib
        self.imports = new_imports

    def __status_definition(self, desc):
        if len(desc) > 1:
            raise ValueError("Experiment description malformed (wrong status definition statement): check your syntax")
        part = desc[0].split(" ")
        if part[0] != 'STATUS':
            raise ValueError("Experiment description malformed (wrong status definition statement): check your syntax")

        self.__statuses[part[1].lower()] = None
        self.model.add_status(part[1])

        return "%s.add_status('%s')\n" % (self.__model_name, part[1])

    def __execution_statement(self, desc):

        if len(desc) > 1:
            raise ValueError("Experiment description malformed (wrong execution statement): check your syntax")
        part = desc[0].split(" ")
        if part[0] != 'EXECUTE' or part[2] != 'ON' or part[4] != 'FOR':
            raise ValueError("Experiment description malformed (wrong execution statement): check your syntax")

        if self.__model_name != part[1]:
            raise ValueError("Execution Definition Error: model '%s' not defined" % part[1])

        if self.__net_name != part[3]:
            raise ValueError("Execution Definition Error: graph '%s' not defined" % part[3])

        return "iterations = %s.iteration_bunch(%s, node_status=False)\n" \
               "res = json.dumps(iterations)\n" \
               "print(res)\n" % (self.__model_name, part[5])

    def __network_generation(self, desc):

        components = {'CREATE_NETWORK': None, 'TYPE': None, 'PARAM': []}

        for part in desc:
            part = part.split(" ")
            if part[0] not in components:
                raise ValueError("Unsupported description")
            if part[0] == 'PARAM':
                if len(part[0]) < 3:
                    raise ValueError("Experiment description malformed (wrong network definition): check your syntax")
                else:
                    components['PARAM'].append((part[1], part[2]))
            else:
                components[part[0]] = part[1]
        self.__net_name = components['CREATE_NETWORK']
        parameters = ""
        for pr in components['PARAM']:
            parameters += "%s=%s, " % (pr[0], pr[1])
        code = "%s = nx.%s(%s)\n" % (self.__net_name, components['TYPE'], parameters)

        return code.replace(", )", ")")

    def __network_loading(self, desc):

        compartments = ['LOAD_NETWORK', 'FROM']

        if len(desc) > 1:
            raise ValueError("Unsupported description")
        stm = desc[0].split(" ")
        if len(stm) != 4 or stm[0] not in compartments or stm[2] not in compartments:
            raise ValueError("Experiment description malformed (wrong network loading statement): check your syntax")

        self.__net_name = stm[1]
        filename = stm[3]

        if os.path.isfile(filename):
            return "%s = nx.read_edgelist('%s')\n" % (self.__net_name, filename)
        else:
            raise ValueError("Experiment description malformed (file not existing): check your syntax")

    def __model_creation(self, desc):

        if len(desc) > 1:
            raise ValueError("Unsupported description")
        self.__model_name = desc[0].split(" ")[1]
        create = "%s = CompositeModel(%s)\n" % (self.__model_name, self.__net_name)
        return create

    def __model_configuration(self, desc):

        components = {'INITIALIZE': None, 'SET': []}

        for part in desc:
            part = part.split(" ")
            if part[0] not in components:
                raise ValueError("Unsupported description")
            if part[0] == 'SET':
                if len(part[0]) < 3:
                    raise ValueError("Experiment description malformed: check your syntax")
                else:
                    components['SET'].append((part[1], part[2]))
        conf = "config = Configuration()\n"
        for cf in components['SET']:
            status = cf[0].lower()
            if status not in self.__statuses:
                raise ValueError("Configuration Error: status not defined")

            conf += "config.add_model_parameter('percentage_%s', %s)\n" % (status, cf[1])
        conf += "%s.set_initial_status(config)\n" % self.__model_name
        return conf

    def __rule_definition(self, desc):

        components = {'RULE': None, 'FROM': None, 'TO': None, 'USING': None}
        for part in desc:
            part = part.split(" ")
            if part[0] not in components:
                raise ValueError("Unsupported description")
            if part[0] != 'RULE':
                if len(part) == 2:
                    components[part[0]] = part[1]
                else:
                    raise ValueError("Unsupported parameter")

        if components['FROM'].lower() not in self.__statuses or components['TO'].lower() not in self.__statuses:
            raise ValueError("Rule Definition Error: status not defined")

        if components['USING'] not in self.__compartments:
            raise ValueError("Conditional Compartment Definition Error: compartment '%s' undefined"
                             % components['USING'])

        apply = "%s.add_rule('%s', '%s', %s)\n" % (self.__model_name, components['FROM'],
                                                   components['TO'], components['USING'])
        return apply

    def __compartment_definition(self, desc):

        components = {'COMPARTMENT': None, 'TYPE': None, 'TRIGGER': None, 'COMPOSE': None,
                      'PARAM': {'probability': 1, 'threshold': None, 'rate': None, 'attribute': None,
                                'attribute_value': None, 'name': None, 'iterations': None}}
        for part in desc:
            part = part.split(" ")
            if part[0] not in components:
                raise ValueError("Unsupported description")
            if len(part) == 2:
                components[part[0]] = part[1]
            else:
                if part[1] not in components['PARAM']:
                    raise ValueError("Unsupported parameter")
                components['PARAM'][part[1]] = part[2]

        if components['TRIGGER'] is not None and components['TRIGGER'].lower() not in self.__statuses:
            raise ValueError("Rule Definition Error: status not defined")

        self.__compartments[components['COMPARTMENT']] = components['TYPE']

        rule = '%s = %s(composed=%s, triggering_status=\'%s\', ' \
               'rate=%s, probability=%s, threshold=%s, attribute=\'%s\', attribute_value=%s, name="%s", iterations=%s)\n' % \
               (components['COMPARTMENT'], components['TYPE'],
                components['COMPOSE'], components['TRIGGER'],
                components['PARAM']['rate'], components['PARAM']['probability'], components['PARAM']['threshold'],
                components['PARAM']['attribute'], components['PARAM']['attribute_value'], components['PARAM']['name'],
                components['PARAM']['iterations'])

        rule = rule.replace("'None'", "None")

        # Code cleaning
        rule = re.sub(', [a-zA-Z\\_]+=None', '', rule)
        rule = rule.replace("  ", " ")
        rule = re.sub('[a-zA-Z\\_]+=None,', '', rule)
        rule = rule.replace("( ", "(")

        return rule

    def __conditional_compartment_composition(self, desc):

        components = ['IF', 'THEN', 'ELSE', 'AS']

        for part in desc:
            part = part.split(" ")
            if part[0] not in components or part[2] not in components \
                    or part[4] not in components or part[6] not in components:
                raise ValueError("Unsupported description")
            if len(part) == 8:

                if part[1] not in self.__compartments or part[3] not in self.__compartments\
                        or part[5] not in self.__compartments:
                    raise ValueError("Conditional Compartment Definition Error: compartment undefined")

                self.__compartments[part[-1]] = "ConditionalComposition"
                return "%s = ConditionalComposition(%s, %s, %s)\n" \
                       % (part[-1], part[1], part[3], part[5])
            else:
                raise ValueError("Experiment description malformed: check your syntax")

    @staticmethod
    def __check_components(identified_directives):

        # Check model
        if 'MODEL' not in identified_directives:
            raise ValueError("Experiment description malformed (Model not specified): check your syntax")

        # Check network
        if 'LOAD_NETWORK' not in identified_directives:
            if 'CREATE_NETWORK' not in identified_directives:
                raise ValueError("Experiment description malformed (Network not specified): check your syntax")
        else:
            if 'CREATE_NETWORK' in identified_directives:
                ValueError("Experiment description malformed (Network not specified): check your syntax")

        # Check execution
        if 'EXECUTE' not in identified_directives:
            raise ValueError("Experiment description malformed (Execution statement missing): check your syntax")

        # Initial status
        if 'INITIALIZE' not in identified_directives:
            raise ValueError("Experiment description malformed (Initial status missing): check your syntax")

    @staticmethod
    def __sanitize_string(text):
        text = text.replace("\t", " ").replace("eval", "").replace("exec", "").replace("__", "")
        return re.sub('[-\\\():=!@#$]', '', text)
